@@ -5,12 +5,53 @@ use strictures 2;
 use autodie;
 
 use version;
+use Log::Any ();
 use Path::Iterator::Rule;
 use Path::Tiny;
 use Fcntl qw/:seek/;
 use Encode;
-use Regexp::Pattern;
-use Regexp::Pattern::License 3.1.94;
+use Regexp::Pattern 0.2.0 (
+	're',
+	'License::*' => (
+		engine   => 'RE2',
+		subject  => 'trait',
+		-prefix  => 'TRAIT_',
+		-has_tag => 'type:trait',
+	),
+	'License::licensed_under' => (
+		subject => 'trait',
+		-prefix => 'TRAIT_GLOBAL_',
+	),
+	'License::version' => (
+		capture => 'numbered',
+		subject => 'trait',
+		-prefix => 'TRAIT_KEEP_',
+	),
+	'License::version_numberstring' => (
+		capture => 'numbered',
+		subject => 'trait',
+		-prefix => 'TRAIT_KEEP_',
+	),
+	'License::*' => (
+		engine     => 'RE2',
+		subject    => 'name',
+		-prefix    => 'NAME_',
+		-lacks_tag => 'type:trait',
+	),
+	'License::*' => (
+		subject    => 'name',
+		-prefix    => 'NAME_GLOBAL_',
+		-lacks_tag => 'type:trait',
+	),
+	'License::*' => (
+		engine     => 'RE2',
+		subject    => 'grant,license',
+		-prefix    => 'GRANT_LICENSE_',
+		-lacks_tag => 'type:trait',
+	),
+);
+
+use Regexp::Pattern::License 3.1.96;
 use String::Copyright 0.003 {
 	format => sub { join ' ', $_->[0] || (), $_->[1] || () }
 };
@@ -21,7 +62,6 @@ use String::Copyright 0.003 {
 	'copyright' => { -as => 'copyright_optimistic' };
 
 use Moo;
-with qw(MooX::Role::Logger);
 
 use experimental "switch";
 use namespace::clean;
@@ -32,11 +72,11 @@ App::Licensecheck - functions for a simple license checker for source files
 
 =head1 VERSION
 
-Version v3.0.38
+Version v3.0.39
 
 =cut
 
-our $VERSION = version->declare("v3.0.38");
+our $VERSION = version->declare("v3.0.39");
 
 =head1 SYNOPSIS
 
@@ -59,6 +99,7 @@ See the script for casual usage.
 # TODO: make naming scheme configurable
 my %L = licensepatterns(qw(debian spdx));
 
+my @RE_NAME              = sort map /^NAME_GLOBAL_(.*)/, keys(%RE);
 my @L_family_cc          = sort keys %{ $L{family}{cc} };
 my @L_type_singleversion = sort keys %{ $L{type}{singleversion} };
 my @L_type_versioned     = sort keys %{ $L{type}{versioned} };
@@ -122,6 +163,11 @@ my $default_ignore_regex = q!
 	\.shelf|_MTN|\.bzr(?:\.backup|tags)?)(?:$|/.*$)
 !;
 
+has log => (
+	is      => 'ro',
+	default => sub { Log::Any->get_logger },
+);
+
 has check_regex => (
 	is     => 'rw',
 	lazy   => 1,
@@ -440,7 +486,6 @@ sub licensepatterns
 	}
 	for my $trait (
 		qw(any_of licensed_under or_at_option
-		clause_advertising clause_non_endorsement clause_reproduction
 		fsf_unlimited fsf_unlimited_retention
 		version_later version_later_postfix)
 		)
@@ -448,64 +493,54 @@ sub licensepatterns
 		my $re = re( "License::$trait", subject => 'trait' );
 		$list{re_trait}{$trait} = ref($re) ? $re : qr/$re/;
 	}
-	for my $trait (qw(version version_numberstring)) {
-		my $re = re(
-			"License::$trait", subject => 'trait',
-			capture => 'numbered'
-		);
-		$list{re_trait_keep}{$trait} = ref($re) ? $re : qr/$re/;
-	}
 
 	#<<<  do not let perltidy touch this (keep long regex on one line)
-	$list{re_grant_license}{local}{version_agpl_gpl}{1} = qr/$list{re_trait_keep}{version_numberstring}(?:[.,])? (?:\(?only\)?.? )?(?:of $list{re_name}{gnu}|(as )?published by the Free Software Foundation)/i;
-	$list{re_grant_license}{local}{version_agpl_gpl}{2} = qr/$list{re_name}{gnu}\b[;,] $list{re_trait_keep}{version_numberstring}\b[.,]? /i;
-	$list{re_grant_license}{local}{version_agpl_gpl}{3} = qr/either $list{re_trait_keep}{version_numberstring},? $list{re_trait}{version_later_postfix}/;
-	$list{re_grant_license}{local}{version_agpl_gpl}{4} = qr/GPL as published by the Free Software Foundation, $list{re_trait_keep}{version_numberstring}/i;
+	$list{re_grant_license}{local}{version_agpl_gpl}{1} = qr/$RE{TRAIT_KEEP_version_numberstring}(?:[.,])? (?:\(?only\)?.? )?(?:of $list{re_name}{gnu}|(as )?published by the Free Software Foundation)/i;
+	$list{re_grant_license}{local}{version_agpl_gpl}{2} = qr/$list{re_name}{gnu}\b[;,] $RE{TRAIT_KEEP_version_numberstring}\b[.,]? /i;
+	$list{re_grant_license}{local}{version_agpl_gpl}{3} = qr/either $RE{TRAIT_KEEP_version_numberstring},? $list{re_trait}{version_later_postfix}/;
+	$list{re_grant_license}{local}{version_agpl_gpl}{4} = qr/GPL as published by the Free Software Foundation, $RE{TRAIT_KEEP_version_numberstring}/i;
 	$list{re_grant_license}{local}{address_agpl_gpl_lgpl}{1} = qr/(?:675 Mass Ave|59 Temple Place|51 Franklin Steet|02139|02111-1307)/i;
 	$list{re_grant_license}{local}{exception_agpl_gpl_lgpl}{1} = qr/permission (?:is (also granted|given))? to link (the code of )?this program with (any edition of )?(Qt|the Qt library)/i;
 	$list{re_grant_license}{local}{generated}{1} = qr/unless such copies or derivative works are solely in the form of machine-executable object code generated by a source language processor/;
 	$list{re_grant_license}{local}{generated}{2} = qr/(All changes made in this file will be lost|DO NOT ((?:HAND )?EDIT|delete this file|modify)|edit the original|Generated (automatically|by|from|data|with)|generated.*file|auto[- ]generated)/i;
-	$list{re_grant_license}{local}{multi}{1} = qr/$list{re_trait}{licensed_under}$list{re_trait}{any_of}(?:[^.]|\.\S)*$list{re_name}{lgpl}$list{re_trait_keep}{version}?/i;
-	$list{re_grant_license}{local}{multi}{2} = qr/$list{re_trait}{licensed_under}$list{re_trait}{any_of}(?:[^.]|\.\S)*$list{re_name}{gpl}$list{re_trait_keep}{version}?/i;
-	$list{re_grant_license}{local}{lgpl}{1} = qr/$list{re_trait}{licensed_under}$list{re_trait}{any_of}(?:[^.]|\.\S)*$list{re_name}{lgpl}$list{re_trait_keep}{version}?/i;
-	$list{re_grant_license}{local}{lgpl}{2} = qr/either $list{re_name}{afl}$list{re_trait_keep}{version}, or $list{re_name}{lgpl}$list{re_trait_keep}{version}?/i;
-	$list{re_grant_license}{local}{lgpl}{3} = qr/either $list{re_name}{gpl}$list{re_trait_keep}{version}?(?: \((?:the )?"?GPL"?\))?, or $list{re_name}{lgpl}/i;
-	$list{re_grant_license}{local}{lgpl}{4} = qr/$list{re_trait}{licensed_under}$list{re_trait_keep}{version}? of $list{re_name}{lgpl}/i;
-	$list{re_grant_license}{local}{lgpl}{5} = qr/$list{re_trait}{licensed_under}$list{re_name}{lgpl}\b[,;:]?(?: either)? ?$list{re_trait_keep}{version_numberstring},? $list{re_trait}{or_at_option} $list{re_trait_keep}{version_numberstring}/i;
-	$list{re_grant_license}{local}{lgpl}{6} = qr/$list{re_trait}{licensed_under}$list{re_name}{lgpl}(?:[,;:]?(?: either)?$list{re_trait_keep}{version}?)?/i;
-	$list{re_grant_license}{local}{agpl}{1} = qr/$list{re_trait}{licensed_under}$list{re_name}{agpl}/i;
+	$list{re_grant_license}{local}{multi}{1} = qr/$RE{TRAIT_GLOBAL_licensed_under}$list{re_trait}{any_of}(?:[^.]|\.\S)*$list{re_name}{lgpl}$RE{TRAIT_KEEP_version}?/i;
+	$list{re_grant_license}{local}{multi}{2} = qr/$RE{TRAIT_GLOBAL_licensed_under}$list{re_trait}{any_of}(?:[^.]|\.\S)*$list{re_name}{gpl}$RE{TRAIT_KEEP_version}?/i;
+	$list{re_grant_license}{local}{lgpl}{1} = qr/$RE{TRAIT_GLOBAL_licensed_under}$list{re_trait}{any_of}(?:[^.]|\.\S)*$list{re_name}{lgpl}$RE{TRAIT_KEEP_version}?/i;
+	$list{re_grant_license}{local}{lgpl}{2} = qr/either $list{re_name}{afl}$RE{TRAIT_KEEP_version}, or $list{re_name}{lgpl}$RE{TRAIT_KEEP_version}?/i;
+	$list{re_grant_license}{local}{lgpl}{3} = qr/either $list{re_name}{gpl}$RE{TRAIT_KEEP_version}?(?: \((?:the )?"?GPL"?\))?, or $list{re_name}{lgpl}/i;
+	$list{re_grant_license}{local}{lgpl}{4} = qr/$RE{TRAIT_GLOBAL_licensed_under}$RE{TRAIT_KEEP_version}? of $list{re_name}{lgpl}/i;
+	$list{re_grant_license}{local}{lgpl}{5} = qr/$RE{TRAIT_GLOBAL_licensed_under}$list{re_name}{lgpl}\b[,;:]?(?: either)? ?$RE{TRAIT_KEEP_version_numberstring},? $list{re_trait}{or_at_option} $RE{TRAIT_KEEP_version_numberstring}/i;
+	$list{re_grant_license}{local}{lgpl}{6} = qr/$RE{TRAIT_GLOBAL_licensed_under}$list{re_name}{lgpl}(?:[,;:]?(?: either)?$RE{TRAIT_KEEP_version}?)?/i;
+	$list{re_grant_license}{local}{agpl}{1} = qr/$RE{TRAIT_GLOBAL_licensed_under}$list{re_name}{agpl}/i;
 	$list{re_grant_license}{local}{agpl}{2} = qr/(?:signe la|means the) GNU Affero General Public License/i;
 	$list{re_grant_license}{local}{agpl}{3} = qr/GNU Affero General Public License into/i;
 	$list{re_grant_license}{local}{agpl}{4} = qr/means either the GNU/i;
-	$list{re_grant_license}{local}{agpl}{5} = qr/AFFERO GENERAL PUBLIC LICENSE$list{re_trait_keep}{version}?/i;
+	$list{re_grant_license}{local}{agpl}{5} = qr/AFFERO GENERAL PUBLIC LICENSE$RE{TRAIT_KEEP_version}?/i;
 	$list{re_grant_license}{local}{gpl}{1} = qr/Copyright \(C\) (?:19|yy)yy <?name of author>?\s+This program is free software/;
 	$list{re_grant_license}{local}{gpl}{2} = qr/under the GNU GPL,? with none of the additional permissions of this License/;
 	$list{re_grant_license}{local}{gpl}{3} = qr/GNU Affero General Public License/;
 	$list{re_grant_license}{local}{gpl}{4} = qr/Terms of the Perl programming language system itself/;
 	$list{re_grant_license}{local}{gpl}{5} = qr/COMPATIBILITY WITH THE GPL LICENSE/i;
-	$list{re_grant_license}{local}{gpl}{6} = qr/$list{re_trait}{licensed_under}$list{re_trait}{any_of}(?:[^.]|\.\S)*$list{re_name}{gpl}$list{re_trait_keep}{version}?/i;
-	$list{re_grant_license}{local}{gpl}{7} = qr/either $list{re_name}{gpl}$list{re_trait_keep}{version}?(?: \((?:the )?"?GPL"?\))?, or $list{re_name}{lgpl}$list{re_trait_keep}{version}?/i;
-	$list{re_grant_license}{local}{gpl}{8} = qr/$list{re_trait}{licensed_under}(?:version \S+ (?:\(?only\)? )?of )?$list{re_name}{gpl}/i;
-	$list{re_grant_license}{local}{gpl}{9} = qr/$list{re_trait}{licensed_under}(?:version \S+ (?:\(?only\)? )?of )?$list{re_name}{gpl}$list{re_trait_keep}{version}?/i;
+	$list{re_grant_license}{local}{gpl}{6} = qr/$RE{TRAIT_GLOBAL_licensed_under}$list{re_trait}{any_of}(?:[^.]|\.\S)*$list{re_name}{gpl}$RE{TRAIT_KEEP_version}?/i;
+	$list{re_grant_license}{local}{gpl}{7} = qr/either $list{re_name}{gpl}$RE{TRAIT_KEEP_version}?(?: \((?:the )?"?GPL"?\))?, or $list{re_name}{lgpl}$RE{TRAIT_KEEP_version}?/i;
+	$list{re_grant_license}{local}{gpl}{8} = qr/$RE{TRAIT_GLOBAL_licensed_under}(?:version \S+ (?:\(?only\)? )?of )?$list{re_name}{gpl}/i;
+	$list{re_grant_license}{local}{gpl}{9} = qr/$RE{TRAIT_GLOBAL_licensed_under}(?:version \S+ (?:\(?only\)? )?of )?$list{re_name}{gpl}$RE{TRAIT_KEEP_version}?/i;
 	foreach my $id ( sort keys %{ $list{family}{cc} } ) {
-		$list{re_grant_license}{local}{cc}{1}{$id} = qr/$list{re_name}{$id}$list{re_trait_keep}{version_numberstring} or $list{re_trait_keep}{version_numberstring}/i;
-		$list{re_grant_license}{local}{cc}{2}{$id} = qr/$list{re_name}{$id}$list{re_trait_keep}{version}(?:,? (?:and|or) (?:the )?(?:GNU )?([AL]?GPL)-?$list{re_trait_keep}{version})?/i;
+		$list{re_grant_license}{local}{cc}{1}{$id} = qr/$list{re_name}{$id}$RE{TRAIT_KEEP_version_numberstring} or $RE{TRAIT_KEEP_version_numberstring}/i;
+		$list{re_grant_license}{local}{cc}{2}{$id} = qr/$list{re_name}{$id}$RE{TRAIT_KEEP_version}(?:,? (?:and|or) (?:the )?(?:GNU )?([AL]?GPL)-?$RE{TRAIT_KEEP_version})?/i;
 	}
 	$list{re_grant_license}{local}{bsd}{1} = qr/THIS SOFTWARE IS PROVIDED .*AS IS AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY/;
-	$list{re_grant_license}{local}{bsd}{2} = qr/$list{re_trait}{clause_advertising}/i;
-	$list{re_grant_license}{local}{bsd}{3} = qr/$list{re_trait}{clause_non_endorsement}/i;
-	$list{re_grant_license}{local}{bsd}{4} = qr/$list{re_trait}{clause_reproduction}/i;
 	$list{re_grant_license}{local}{bsd}{5} = qr/licen[sc]e:? ?bsd-(\d)-clause/i;
 	$list{re_grant_license}{local}{bsd}{6} = qr/licen[sc]e:? ?bsd\b/i;
-	$list{re_grant_license}{local}{apache}{1} = qr/$list{re_name}{apache}$list{re_trait_keep}{version}?(?:(?: or)? [^ ,]*?apache[^ ,]*| \([^(),]+\))*,? or $list{re_name}{gpl}$list{re_trait_keep}{version}?/i;
-	$list{re_grant_license}{local}{apache}{2} = qr/$list{re_name}{apache}$list{re_trait_keep}{version}?(?:(?: or)? [^ ,]*?apache[^ ,]*| \([^(),]\))*,? or(?: the)? bsd(?:[ -](\d)-clause)?\b/i;
-	$list{re_grant_license}{local}{apache}{4} = qr/$list{re_name}{apache}$list{re_trait_keep}{version}?(?:(?: or)? [^ ,]*?apache[^ ,]*| \([^(),]\))*,? or $list{re_name}{mit}\b/i;
+	$list{re_grant_license}{local}{apache}{1} = qr/$list{re_name}{apache}$RE{TRAIT_KEEP_version}?(?:(?: or)? [^ ,]*?apache[^ ,]*| \([^(),]+\))*,? or $list{re_name}{gpl}$RE{TRAIT_KEEP_version}?/i;
+	$list{re_grant_license}{local}{apache}{2} = qr/$list{re_name}{apache}$RE{TRAIT_KEEP_version}?(?:(?: or)? [^ ,]*?apache[^ ,]*| \([^(),]\))*,? or(?: the)? bsd(?:[ -](\d)-clause)?\b/i;
+	$list{re_grant_license}{local}{apache}{4} = qr/$list{re_name}{apache}$RE{TRAIT_KEEP_version}?(?:(?: or)? [^ ,]*?apache[^ ,]*| \([^(),]\))*,? or $list{re_name}{mit}\b/i;
 	$list{re_grant_license}{local}{fsful}{1} = qr/This (\w+)(?: (?:file|script))? is free software; $list{re_trait}{fsf_unlimited}/i;
 	$list{re_grant_license}{local}{fsfullr}{1} = qr/This (\w+)(?: (?:file|script))?  is free software; $list{re_trait}{fsf_unlimited_retention}/i;
-	$list{re_grant_license}{local}{php}{1} = qr/$list{re_trait}{licensed_under}$list{re_trait_keep}{version_numberstring} of the PHP license/;
+	$list{re_grant_license}{local}{php}{1} = qr/$RE{TRAIT_GLOBAL_licensed_under}$RE{TRAIT_KEEP_version_numberstring} of the PHP license/;
 	$list{re_grant_license}{local}{pd}{1} = qr/$list{re_grant_license}{public_domain}/;
 	foreach my $id ( sort keys %{ $list{type}{versioned} } ) {
-		$list{re_grant_license}{local}{versioned}{$id} = qr/$list{re_name}{$id}$list{re_trait_keep}{version}?/;
+		$list{re_grant_license}{local}{versioned}{$id} = qr/$list{re_name}{$id}$RE{TRAIT_KEEP_version}?/;
 	}
 	$list{re_grant_license}{local}{trailing_space} = qr/\s+$/;
 	#>>>
@@ -513,17 +548,6 @@ sub licensepatterns
 	return %list;
 }
 
-sub _log_license
-{
-	my ( $self, $action, $thing, $extra ) = @_;
-	if ( $self->_logger->is_trace() ) {
-		$self->_logger->trace( "license text $action: $thing", ($extra) );
-	}
-	else {
-		$self->_logger->debug("license text $action: $thing");
-	}
-}
-
 sub parse_license
 {
 	my ( $self, $licensetext, $file, $position ) = @_;
@@ -589,19 +613,31 @@ sub parse_license
 			$v2    ? "(v$v2)"    : (),
 		);
 		my $expr = join( ' or ', @spdx );
-		$self->_log_license( 'collected', $expr, { path => "$file" } )
-			if ( $self->_logger->is_debug() );
+		$self->log->debugf(
+			'collected license expression: %s [%s]',
+			$expr, $file
+		);
 		push @spdx_license, $expr;
 		$license = join( ' ', $L{caption}{$legacy} || $legacy, $license );
 	};
 
 	for my $trait (qw(licensed_under)) {
-		while ( $licensetext =~ /$L{re_trait}{$trait}/g ) {
+		next unless ( $licensetext =~ /$RE{"TRAIT_$trait"}/ );
+		while ( $licensetext =~ /$RE{"TRAIT_GLOBAL_$trait"}/g ) {
+			$self->log->tracef(
+				'located license trait "%s": %s [%s]',
+				$trait, $-[0], $file
+			);
 			$match{_}{$trait}{ $-[0] } = $+[0];
 		}
 	}
-	for my $id ( keys %{ $L{re_name} } ) {
-		while ( $licensetext =~ /$L{re_name}{$id}/g ) {
+	for my $id (@RE_NAME) {
+		next unless ( $licensetext =~ /$RE{"NAME_$id"}/ );
+		while ( $licensetext =~ /$RE{"NAME_GLOBAL_$id"}/g ) {
+			$self->log->tracef(
+				'located license name: %s: %s [%s]',
+				$id, $-[0], $file
+			);
 			$match{$id}{name}{ $-[0] } = $+[0];
 		}
 	}
@@ -630,16 +666,28 @@ sub parse_license
 			# may include version
 			if ( grep { $_ eq $name } @L_type_versioned ) {
 				if (substr( $licensetext, $pos2 )
-					=~ /^$L{re_trait_keep}{version}/ )
+					=~ /^$RE{TRAIT_KEEP_version}/ )
 				{
+					$self->log->tracef(
+						'located license version: %s: %s [%s]',
+						$1, $pos2, $file
+					);
 					$pos2 += $+[0];
 					my $version = $1;
 					$version =~ s/(?:\.0)+$//;
 					$_ .= "_$version";
 					if ($2) {
+						$self->log->tracef(
+							'located license trait "%s": %s: %s [%s]',
+							'or_later', $2, $-[0], $file
+						);
 						$_ .= "_or_later";
 					}
 				}
+				$self->log->tracef(
+					'located license grant: %s: %s [%s]',
+					$_, $pos, $file
+				);
 				$pos{$pos}{$_} = 1;
 			}
 		}
@@ -647,6 +695,10 @@ sub parse_license
 	for my $pos ( keys %pos ) {
 		my @name = keys %{ $pos{$pos} };
 		for my $name (@name) {
+			$self->log->tracef(
+				'flagged license grant: %s: %s [%s]',
+				$name, $pos, $file
+			);
 			$grant{$name} = 1;
 		}
 	}
@@ -656,18 +708,34 @@ sub parse_license
 		# version of AGPL/GPL
 		given ($licensetext) {
 			when ( $L{re_grant_license}{local}{version_agpl_gpl}{1} ) {
+				$self->log->tracef(
+					'detected custom pattern version_agpl_gpl#1: %s: %s [%s]',
+					$1, $-[0], $file
+				);
 				$gplver      = " (v$1)";
 				@spdx_gplver = ($1)
 			}
 			when ( $L{re_grant_license}{local}{version_agpl_gpl}{2} ) {
+				$self->log->tracef(
+					'detected custom pattern version_agpl_gpl#2: %s: %s [%s]',
+					$1, $-[0], $file
+				);
 				$gplver      = " (v$1)";
 				@spdx_gplver = ($1);
 			}
 			when ( $L{re_grant_license}{local}{version_agpl_gpl}{3} ) {
+				$self->log->tracef(
+					'detected custom pattern version_agpl_gpl#3: %s: %s [%s]',
+					$1, $-[0], $file
+				);
 				$gplver      = " (v$1 or later)";
 				@spdx_gplver = ( $1 . '+' );
 			}
 			when ( $L{re_grant_license}{local}{version_agpl_gpl}{4} ) {
+				$self->log->tracef(
+					'detected custom pattern version_agpl_gpl#4: %s: %s [%s]',
+					$1, $-[0], $file
+				);
 				$gplver      = " (v$1)";
 				@spdx_gplver = ($1)
 			}
@@ -679,6 +747,10 @@ sub parse_license
 		# address in AGPL/GPL/LGPL
 		given ($licensetext) {
 			when ( $L{re_grant_license}{local}{address_agpl_gpl_lgpl}{1} ) {
+				$self->log->tracef(
+					'detected custom pattern address_agpl_gpl_lgpl#1: %s [%s]',
+					$-[0], $file
+				);
 				$extrainfo = " (with incorrect FSF address)$extrainfo";
 			}
 		}
@@ -686,6 +758,10 @@ sub parse_license
 		# exception for AGPL/GPL/LGPL
 		given ($licensetext) {
 			when ( $L{re_grant_license}{local}{exception_agpl_gpl_lgpl}{1} ) {
+				$self->log->tracef(
+					'detected custom pattern exception_agpl_gpl_lgpl#1: %s [%s]',
+					$-[0], $file
+				);
 				$extrainfo  = " (with Qt exception)$extrainfo";
 				$spdx_extra = 'with Qt exception';
 			}
@@ -697,9 +773,17 @@ sub parse_license
 
 		# exclude blurb found in boost license text
 		when ( $L{re_grant_license}{local}{generated}{1} ) {
+			$self->log->tracef(
+				'detected and skipped custom pattern generated#1: %s [%s]',
+				$-[0], $file
+			);
 			break;
 		}
 		when ( $L{re_grant_license}{local}{generated}{2} ) {
+			$self->log->tracef(
+				'detected custom pattern generated#2: %s [%s]',
+				$-[0], $file
+			);
 			$license = "GENERATED FILE";
 		}
 	}
@@ -710,11 +794,19 @@ sub parse_license
 	# same sentence
 	if ( grep { $match{$_}{name} } @lgpl ) {
 		if ( $licensetext =~ $L{re_grant_license}{local}{multi}{1} ) {
+			$self->log->tracef(
+				'detected custom pattern multi#1: %s %s %s: %s [%s]',
+				'lgpl', $1, $2, $-[0], $file
+			);
 			push @multilicenses, 'lgpl', $1, $2;
 		}
 	}
 	if ( grep { $match{$_}{name} } @gpl ) {
 		if ( $licensetext =~ $L{re_grant_license}{local}{multi}{2} ) {
+			$self->log->tracef(
+				'detected custom pattern multi#2: %s %s %s: %s [%s]',
+				'gpl', $1, $2, $-[0], $file
+			);
 			push @multilicenses, 'gpl', $1, $2;
 		}
 	}
@@ -727,35 +819,65 @@ sub parse_license
 
 			# LGPL, among several
 			when ( $L{re_grant_license}{local}{lgpl}{1} ) {
+				$self->log->tracef(
+					'detected and skipped custom pattern lgpl#1: %s [%s]',
+					$-[0], $file
+				);
 				break;    # handled in multi-licensing loop
 			}
 
 			# AFL or LGPL
 			when ( $L{re_grant_license}{local}{lgpl}{2} ) {
+				$self->log->tracef(
+					'detected and skipped custom pattern lgpl#2: %s [%s]',
+					$-[0], $file
+				);
 				break;    # handled in AFL loop
 			}
 
 			# GPL or LGPL
 			when ( $L{re_grant_license}{local}{lgpl}{3} ) {
+				$self->log->tracef(
+					'detected and skipped custom pattern lgpl#3: %s [%s]',
+					$-[0], $file
+				);
 				break;    # handled in GPL loop
 			}
 
 			# LGPL, version first
 			when ( $L{re_grant_license}{local}{lgpl}{4} ) {
+				$self->log->tracef(
+					'detected custom pattern lgpl#4: %s %s %s: %s [%s]',
+					'lgpl', $1, $2, $-[0], $file
+				);
 				$gen_license->( 'lgpl', $1, $2 );
 			}
 
 			# LGPL, dual versions last
 			when ( $L{re_grant_license}{local}{lgpl}{5} ) {
+				$self->log->tracef(
+					'detected custom pattern lgpl#5: %s %s %s: %s [%s]',
+					'lgpl', $1, $2, $-[0], $file
+				);
 				$license = "LGPL (v$1 or v$2) $license";
-				push @spdx_license, "LGPL-$1 or LGPL-$2";
+				my $expr = "LGPL-$1 or LGPL-$2";
+				$self->log->debugf(
+					'collected license expression: %s [%s]',
+					$expr, $file
+				);
+				push @spdx_license, $expr;
 			}
 
 			# LGPL, version last
 			when ( $L{re_grant_license}{local}{lgpl}{6} ) {
+				$self->log->tracef(
+					'detected custom pattern lgpl#6: %s %s %s: %s [%s]',
+					'lgpl', $1, $2, $-[0], $file
+				);
 				$gen_license->( 'lgpl', $1, $2 );
 			}
 		}
+		$self->log->tracef( 'flagged license objects: lgpl afl [%s]', $file );
 		$match{lgpl}{custom} = 1;
 		$match{afl}{custom}  = 1;
 	}
@@ -765,28 +887,54 @@ sub parse_license
 		# AGPL
 		given ($licensetext) {
 			when ( $L{re_grant_license}{local}{agpl}{1} ) {
+				$self->log->tracef(
+					'detected custom pattern agpl#1: %s: %s [%s]',
+					'agpl', $-[0], $file
+				);
 				$license = "AGPL$gplver$extrainfo $license";
-				push @spdx_license, $gen_spdx->('AGPL');
+				my @expr = $gen_spdx->('AGPL');
+				$self->log->debugf(
+					'collected license expression: %s [%s]',
+					[@expr], $file
+				);
+				push @spdx_license, @expr;
 			}
 
 			# exclude CeCILL-2.1 license
 			when ( $L{re_grant_license}{local}{agpl}{2} ) {
+				$self->log->tracef(
+					'detected and skipped custom pattern agpl#2: %s [%s]',
+					$-[0], $file
+				);
 				break;
 			}
 
 			# exclude GPL-3 license
 			when ( $L{re_grant_license}{local}{agpl}{3} ) {
+				$self->log->tracef(
+					'detected and skipped custom pattern agpl#3: %s [%s]',
+					$-[0], $file
+				);
 				break;
 			}
 
 			# exclude MPL-2.0 license
 			when ( $L{re_grant_license}{local}{agpl}{4} ) {
+				$self->log->tracef(
+					'detected and skipped custom pattern agpl#4: %s [%s]',
+					$-[0], $file
+				);
 				break;
 			}
 			when ( $L{re_grant_license}{local}{agpl}{5} ) {
+				$self->log->tracef(
+					'detected custom pattern agpl#6: agpl %s %s: %s [%s]',
+					$1, $2, $-[0], $file
+				);
 				$gen_license->( 'agpl', $1, $2 );
 			}
 		}
+		$self->log->tracef( 'flagged license object: agpl [%s]', $file );
 		$match{agpl}{custom} = 1;
 	}
 
@@ -797,53 +945,90 @@ sub parse_license
 
 			# exclude GPL fulltext (rarely if ever used as grant)
 			when ( $L{re_grant_license}{local}{gpl}{1} ) {
+				$self->log->tracef(
+					'detected and skipped custom pattern gpl#1: %s [%s]',
+					$-[0], $file
+				);
 				break;
 			}
 
 			# exclude LGPL-3 fulltext (rarely if ever used as grant)
 			when ( $L{re_grant_license}{local}{gpl}{2} ) {
+				$self->log->tracef(
+					'detected and skipped custom pattern gpl#2: %s [%s]',
+					$-[0], $file
+				);
 				break;
 			}
 
 			# exclude AGPL-3 license
 			when ( $L{re_grant_license}{local}{gpl}{3} ) {
+				$self->log->tracef(
+					'detected and skipped custom pattern gpl#3: %s [%s]',
+					$-[0], $file
+				);
 				break;
 			}
 
 			# exclude Perl combo license
 			when ( $L{re_grant_license}{local}{gpl}{4} ) {
+				$self->log->tracef(
+					'detected and skipped custom pattern gpl#4: %s [%s]',
+					$-[0], $file
+				);
 				break;
 			}
 
 			# exclude CeCILL-1.1 license
 			when ( $L{re_grant_license}{local}{gpl}{5} ) {
+				$self->log->tracef(
+					'detected and skipped custom pattern gpl#5: %s [%s]',
+					$-[0], $file
+				);
 				break;
 			}
 
 			# GPL, among several
 			when ( $L{re_grant_license}{local}{gpl}{6} ) {
+				$self->log->tracef(
+					'detected and skipped custom pattern gpl#6: %s [%s]',
+					$-[0], $file
+				);
 				break;    # handled in multi-licensing loop
 			}
 
 			# GPL or LGPL
 			when ( $L{re_grant_license}{local}{gpl}{7} ) {
+				$self->log->tracef(
+					'detected custom pattern agpl#7: %s %s %s %s %s %s %s %s: %s [%s]',
+					'gpl', $1, $2, 'lgpl', $3, $4, $-[0], $file
+				);
 				$gen_license->( 'gpl', $1, $2, 'lgpl', $3, $4 );
 			}
 			if ( $gplver or $extrainfo ) {
 				when ( $L{re_grant_license}{local}{gpl}{8} ) {
+					$self->log->tracef(
+						'detected custom pattern gpl#8: %s: %s [%s]',
+						'gpl', $-[0], $file
+					);
 					$license = "GPL$gplver$extrainfo $license";
 					my $expr = join ' ', $gen_spdx->('GPL');
-					$self->_log_license(
-						'collected', $expr,
-						{ path => "$file" }
-					) if ( $self->_logger->is_debug() );
+					$self->log->debugf(
+						'collected license expression: %s [%s]',
+						$expr, $file
+					);
 					push @spdx_license, $expr;
 				}
 			}
 			when ( $L{re_grant_license}{local}{gpl}{9} ) {
+				$self->log->tracef(
+					'detected custom pattern gpl#9: %s %s %s: %s [%s]',
+					'gpl', $1, $2, $-[0], $file
+				);
 				$gen_license->( 'gpl', $1, $2 );
 			}
 		}
+		$self->log->tracef( 'flagged license object: gpl [%s]', $file );
 		$match{gpl}{custom} = 1;
 	}
 
@@ -852,32 +1037,63 @@ sub parse_license
 		foreach my $id (@L_family_cc) {
 			next unless ( $match{$id}{name} );
 			when ( $L{re_grant_license}{local}{cc}{1}{$id} ) {
+				$self->log->tracef(
+					'detected custom pattern cc#1: %s %s %s: %s [%s]',
+					$id, $1, $2, $-[0], $file
+				);
 				$license = "$L{caption}{$id} (v$1 or v$2) $license";
 				my $expr = "$L{name}{$id}-$1 or $L{name}{$id}-$2";
-				$self->_log_license( 'collected', $expr, { path => "$file" } )
-					if ( $self->_logger->is_debug() );
+				$self->log->debugf(
+					'collected license expression: %s [%s]',
+					$expr, $file
+				);
 				push @spdx_license, $expr;
 			}
 			when ( $L{re_grant_license}{local}{cc}{2}{$id} ) {
+				$self->log->tracef(
+					'detected custom pattern cc#2: %s %s %s %s %s: %s [%s]',
+					$id, $1, $2, $3, $4, $-[0], $file
+				);
 				$gen_license->( $id, $1, $2, $3, $4 );
 			}
 		}
 	}
-	$match{$_}{custom} = 1
-		for (
-		qw(cc_by cc_by_nc cc_by_nc_nd cc_by_nc_sa cc_by_nd cc_by_sa cc_cc0 cc_sp)
-		);
+	for my $id (
+		qw(cc_by cc_by_nc cc_by_nc_nd cc_by_nc_sa cc_by_nd cc_by_sa cc_cc0 cc_nc cc_nd cc_sa cc_sp)
+		)
+	{
+		$match{$id}{custom} = 1;
+		for ( $L{members}{$id} ? @{ $L{members}{$id} } : () ) {
+			$match{$_}{custom} = 1;
+		}
+	}
 
 	# BSD
 	if ( $licensetext =~ $L{re_grant_license}{local}{bsd}{1} ) {
+		$self->log->tracef(
+			'detected custom pattern bsd#1: %s [%s]', $-[0],
+			$file
+		);
 		given ($licensetext) {
-			when ( $L{re_grant_license}{local}{bsd}{2} ) {
+			when ( $RE{TRAIT_clause_advertising} ) {
+				$self->log->tracef(
+					'detected custom pattern bsd#2: %s: %s [%s]',
+					'bsd_4_clause', $-[0], $file
+				);
 				$gen_license->('bsd_4_clause');
 			}
-			when ( $L{re_grant_license}{local}{bsd}{3} ) {
+			when ( $RE{TRAIT_clause_non_endorsement} ) {
+				$self->log->tracef(
+					'detected custom pattern bsd#3: %s: %s [%s]',
+					'bsd_3_clause', $-[0], $file
+				);
 				$gen_license->('bsd_3_clause');
 			}
-			when ( $L{re_grant_license}{local}{bsd}{4} ) {
+			when ( $RE{TRAIT_clause_reproduction} ) {
+				$self->log->tracef(
+					'detected custom pattern bsd#4: %s: %s [%s]',
+					'bsd_2_clause', $-[0], $file
+				);
 				$gen_license->('bsd_2_clause');
 			}
 			default {
@@ -886,63 +1102,114 @@ sub parse_license
 		}
 	}
 	elsif ( $licensetext =~ $L{re_grant_license}{local}{bsd}{5} ) {
+		$self->log->tracef(
+			'detected custom pattern bsd#5: bsd_%s_clause: %s [%s]',
+			$1, $-[0], $file
+		);
 		$gen_license->("bsd_${1}_clause");
 	}
 	elsif ( $licensetext =~ $L{re_grant_license}{local}{bsd}{6} ) {
+		$self->log->tracef(
+			'detected custom pattern bsd#6: %s: %s [%s]',
+			'bsd', $-[0], $file
+		);
 		$gen_license->('bsd');
 	}
+	$self->log->tracef(
+		'flagged license objects: bsd_2_clause bsd_3_clause bsd_4_clause [%s]',
+		$file
+	);
 	$match{$_}{custom} = 1 for (qw(bsd_2_clause bsd_3_clause bsd_4_clause));
 
 	# Apache
 	given ($licensetext) {
 		if ( $match{apache}{name} ) {
 			when ( $L{re_grant_license}{local}{apache}{1} ) {
+				$self->log->tracef(
+					'detected custom pattern apache#1: %s %s %s %s %s %s: %s [%s]',
+					'apache', $1, $2, 'gpl', $3, $4, $-[0], $file
+				);
 				$gen_license->( 'apache', $1, $2, 'gpl', $3, $4 );
 			}
 			when ( $L{re_grant_license}{local}{apache}{2} ) {
+				$self->log->tracef(
+					'detected custom pattern apache#2: %s %s %s bsd_%s_clause: %s [%s]',
+					'apache', $1, $2, $3, $-[0], $file
+				);
 				$gen_license->( 'apache', $1, $2, "bsd_${3}_clause" );
 			}
 			when ( $L{re_grant_license}{local}{apache}{4} ) {
+				$self->log->tracef(
+					'detected custom pattern apache#4: %s %s %s: %s [%s]',
+					'mit', $3, $4, $-[0], $file
+				);
 				$gen_license->( 'mit', $3, $4 );
 			}
 		}
 	}
+	$self->log->tracef( 'flagged license object: apache [%s]', $file );
 	$match{apache}{custom} = 1;
 
 	# FSFUL
 	given ($licensetext) {
-		when ( $L{re_grant_license}{fsful} ) {
+		when ( $RE{GRANT_LICENSE_fsful} ) {
+			$self->log->tracef(
+				'detected grant/license: %s: %s [%s]',
+				'fsful', $-[0], $file
+			);
 			$gen_license->('fsful');
 		}
 		when ( $L{re_grant_license}{local}{fsful}{1} ) {
+			$self->log->tracef(
+				'collected custom pattern fsful#1: %s: %s [%s]',
+				'fsful', $-[0], $file
+			);
 			$license = "FSF Unlimited ($1 derivation) $license";
 			my $expr = "FSFUL~$1";
-			$self->_log_license( 'collected', $expr, { path => "$file" } )
-				if ( $self->_logger->is_debug() );
+			$self->log->debugf(
+				'collected license expression: %s [%s]',
+				$expr, $file
+			);
 			push @spdx_license, $expr;
 		}
 	}
+	$self->log->tracef( 'flagged license object: fsful [%s]', $file );
 	$match{fsful}{custom} = 1;
 
 	# FSFULLR
 	given ($licensetext) {
-		when ( $L{re_grant_license}{fsfullr} ) {
+		when ( $RE{GRANT_LICENSE_fsfullr} ) {
+			$self->log->tracef(
+				'detected grant/license: %s: %s [%s]',
+				'fsfullr', $-[0], $file
+			);
 			$gen_license->('fsfullr');
 		}
 		when ( $L{re_grant_license}{local}{fsfullr}{1} ) {
+			$self->log->tracef(
+				'collected custom pattern fsfullr#1: %s: %s [%s]',
+				'fsfullr', $-[0], $file
+			);
 			$license
 				= "FSF Unlimited (with Retention, $1 derivation) $license";
 			my $expr = "FSFULLR~$1";
-			$self->_log_license( 'collected', $expr, { path => "$file" } )
-				if ( $self->_logger->is_debug() );
+			$self->log->debugf(
+				'collected license expression: %s [%s]',
+				$expr, $file
+			);
 			push @spdx_license, $expr;
 		}
 	}
+	$self->log->tracef( 'flagged license object: fsfullr [%s]', $file );
 	$match{fsfullr}{custom} = 1;
 
 	# JSON
 	given ($licensetext) {
-		when ( $L{re_grant_license}{json} ) {
+		when ( $RE{GRANT_LICENSE_json} ) {
+			$self->log->tracef(
+				'detected grant/license: %s: %s [%s]',
+				'json', $-[0], $file
+			);
 			$gen_license->('JSON');
 		}
 	}
@@ -950,6 +1217,10 @@ sub parse_license
 	# PHP
 	given ($licensetext) {
 		when ( $L{re_grant_license}{local}{php}{1} ) {
+			$self->log->tracef(
+				'detected custom pattern php#1: %s: %s [%s]',
+				'PHP', $-[0], $file
+			);
 			$gen_license->( 'PHP', $1 );
 		}
 	}
@@ -957,9 +1228,14 @@ sub parse_license
 	# public-domain
 	given ($licensetext) {
 		when ( $L{re_grant_license}{local}{pd}{1} ) {
+			$self->log->tracef(
+				'detected custom pattern pd#1: %s: %s [%s]',
+				'public_domain', $-[0], $file
+			);
 			$gen_license->('public_domain');
 		}
 	}
+	$self->log->tracef( 'flagged license object: public_domain [%s]', $file );
 	$match{public_domain}{custom} = 1;
 
 	# singleversion
@@ -967,7 +1243,11 @@ sub parse_license
 		next if ( $match{$id}{custom} );
 
 		if ( grep { $_ ne $id } @L_tidy ) {
-			if ( $licensetext =~ $L{re_grant_license}{$id} ) {
+			if ( $licensetext =~ $RE{"GRANT_LICENSE_$id"} ) {
+				$self->log->tracef(
+					'detected and flagged grant/license: %s [%s]',
+					$id, $-[0], $file
+				);
 				$grant{$id} = 1;
 			}
 		}
@@ -976,8 +1256,13 @@ sub parse_license
 			$gen_license->($id);
 
 			# skip unversioned equivalent
-			$match{ $L{series}{$id} }{custom} = 1
-				if ( $L{series}{$id} );
+			if ( $L{series}{$id} ) {
+				$self->log->tracef(
+					'flagged license object: %s [%s]',
+					$id, $file
+				);
+				$match{ $L{series}{$id} }{custom} = 1;
+			}
 		}
 	}
 
@@ -988,20 +1273,34 @@ sub parse_license
 
 		# skip embedded or referenced licenses
 		if ( grep { $id eq $_ } qw(mpl python) ) {
-			next if $licensetext =~ $L{re_grant_license}{rpsl};
+			if ( $licensetext =~ $RE{GRANT_LICENSE_rpsl} ) {
+				$self->log->tracef(
+					'detected and skipped grant/license: %s [%s]',
+					'rpsl', $-[0], $file
+				);
+				next;
+			}
 		}
 
 		# FIXME: match grant (not name)
 		if ( $match{$id}{name} ) {
 			if ( $licensetext =~ $L{re_grant_license}{local}{versioned}{$id} )
 			{
+				$self->log->tracef(
+					'detected and flagged custom pattern versioned: %s %s %s: [%s]',
+					$id, $1, $2, $-[0], $file
+				);
 				$gen_license->( $id, $1, $2 );
 				$match{$id}{custom} = 1;
 			}
 		}
 		next if ( $match{$id}{custom} );
-		if ( $L{re_grant_license}{$id} ) {
-			if ( $licensetext =~ $L{re_grant_license}{$id} ) {
+		if ( $RE{"GRANT_LICENSE_$id"} ) {
+			if ( $licensetext =~ $RE{"GRANT_LICENSE_$id"} ) {
+				$self->log->tracef(
+					'detected grant/license: %s: [%s]',
+					$id, $-[0], $file
+				);
 				$gen_license->($id);
 			}
 		}
@@ -1014,17 +1313,45 @@ sub parse_license
 
 		# skip embedded or referenced licenses
 		if ( grep { $id eq $_ } qw(zlib) ) {
-			next if $licensetext =~ $L{re_grant_license}{cube};
+			if ( $licensetext =~ $RE{"GRANT_LICENSE_cube"} ) {
+				$self->log->tracef(
+					'skipped grant/license: %s: [%s]',
+					'cube', $-[0], $file
+				);
+				next;
+			}
 		}
 		if ( grep { $id eq $_ } qw(ntp) ) {
-			next if $licensetext =~ $L{re_grant_license}{ntp_disclaimer};
-			next if $licensetext =~ $L{re_grant_license}{dsdp};
+			if ( $licensetext =~ $RE{"GRANT_LICENSE_ntp_disclaimer"} ) {
+				$self->log->tracef(
+					'skipped grant/license: %s: [%s]',
+					'ntp_disclaimer', $-[0], $file
+				);
+				next;
+			}
+			if ( $licensetext =~ $RE{"GRANT_LICENSE_dsdp"} ) {
+				$self->log->tracef(
+					'skipped grant/license: %s: [%s]',
+					'dsdp', $-[0], $file
+				);
+				next;
+			}
 		}
 		if ( grep { $id eq $_ } qw(ntp_disclaimer) ) {
-			next if $licensetext =~ $L{re_grant_license}{mit_cmu};
+			if ( $licensetext =~ $RE{"GRANT_LICENSE_mit_cmu"} ) {
+				$self->log->tracef(
+					'skipped grant/license: %s: [%s]',
+					'mit_cmu', $-[0], $file
+				);
+				next;
+			}
 		}
 
-		if ( $licensetext =~ $L{re_grant_license}{$id} ) {
+		if ( $licensetext =~ $RE{"GRANT_LICENSE_$id"} ) {
+			$self->log->tracef(
+				'detected grant/license: %s: [%s]',
+				$id, $-[0], $file
+			);
 			$gen_license->($id);
 		}
 	}
@@ -1032,8 +1359,10 @@ sub parse_license
 	# Remove trailing spaces.
 	$license =~ s/$L{re_grant_license}{local}{trailing_space}//;
 	my $expr = join( ' and/or ', sort @spdx_license );
-	$self->_log_license( 'resolved', $expr, { path => "$file" } )
-		if ( $self->_logger->is_debug() );
+	$self->log->infof(
+		'resolved license expression: %s [%s]', $expr,
+		$file
+	);
 	return ( $self->deb_fmt ? $expr : $license ) || 'UNKNOWN';
 }
 
